#include <cuda.h>
#include <stdio.h>
#include <stdlib.h>
#include <cmath>
#include <iostream>
#include "MersenneTwister.h"
using namespace std;

MTRand rand1;

#define INFINITY 10 //10 steps is good enough for the Blancmange Function

__device__ double function(double x)
{
	double intpart, fractpart;
	double y = 0;
	fractpart = modf(x, &intpart);

	if(fractpart >= 0.5)
	{
		for(int n = 0; n < INFINITY; n++)
		{
			double power = pow(2.0,n);
			y+= (ceil(power*x) - power*x)/power;
		}
	}
	else
	{
		for(int n = 0; n < INFINITY; n++)
		{
			double power = pow(2.0,n);
			y+= (power*x - floor(power*x)) / power;
		}
	}
	return y;
}

double Hfunction(double x)
{
	double intpart, fractpart;
	double y = 0;
	fractpart = modf(x, &intpart);

	if(fractpart >= 0.5)
	{
		for(int n = 0; n < INFINITY; n++)
		{
			double power = pow(2.0,n);
			y+= (ceil(power*x) - power*x)/power;
		}
	}
	else
	{
		for(int n = 0; n < INFINITY; n++)
		{
			double power = pow(2.0,n);
			y+= (power*x - floor(power*x)) / power;
		}
	}
	return y;
}

__global__ void monteCarloStep(double* inner_points, double* randXs, double* randYs)
{
	if(randYs[threadIdx.x + blockIdx.x*blockDim.x] <= function(randXs[threadIdx.x + blockIdx.x*blockDim.x]))
		inner_points[threadIdx.x + blockIdx.x*blockDim.x] = 1;
}

int main()
{
	const double circleArea = 0.19635; //A = pi * r^2, r = 0.25
	const double r = 0.25;
	const int monte_carlo_steps = 5120000;
	const double x_boundary = 0.5; //generate x from 0 to 0.5
	const double y_boundary_low = 0.25;
	const double y_boundary_high = 0.75; //generate y from 0.25 to 0.75

	double* randXs = (double*)malloc(sizeof(double)*monte_carlo_steps);
	double* randYs = (double*)malloc(sizeof(double)*monte_carlo_steps);
	double* cudaRandXs;
	double* cudaRandYs;

	for(int i = 0; i < monte_carlo_steps; i++)
	{
		double tempRandX = rand1.rand(x_boundary);
		double tempRandY = y_boundary_low + rand1.rand(y_boundary_high - y_boundary_low);
		
		if((tempRandX-0.25)*(tempRandX-0.25) + (tempRandY-0.5)*(tempRandY-0.5) <= (r*r))
		{ //Make sure random number is within the circle
			randXs[i] = tempRandX;
			randYs[i] = tempRandY;
		}
		else
		 	i--;
	}

	double* inner_points = (double*)calloc(monte_carlo_steps, sizeof(double));
	double* cuda_inner_points;
	cudaMalloc(&cuda_inner_points, monte_carlo_steps*sizeof(double));
	cudaMemcpy(cuda_inner_points, inner_points, monte_carlo_steps*sizeof(double), cudaMemcpyHostToDevice);

	cudaMalloc(&cudaRandXs, sizeof(double)*monte_carlo_steps);
	cudaMalloc(&cudaRandYs, sizeof(double)*monte_carlo_steps);
	cudaMemcpy(cudaRandXs, randXs, sizeof(double)*monte_carlo_steps, cudaMemcpyHostToDevice);
	cudaMemcpy(cudaRandYs, randYs, sizeof(double)*monte_carlo_steps, cudaMemcpyHostToDevice);

	monteCarloStep<<<monte_carlo_steps/1024,1024>>>(cuda_inner_points, cudaRandXs, cudaRandYs);

	cudaMemcpy(inner_points, cuda_inner_points, monte_carlo_steps*sizeof(double), cudaMemcpyDeviceToHost);
	long double I = 0;
	for(int q = 0; q < monte_carlo_steps; q++)
		if(inner_points[q] == 1)
			I++;

	long double estimate = (I/monte_carlo_steps) * circleArea;

	cout << "Monte-Carlo Estimate: " << estimate << endl;

	free(inner_points);
	cudaFree(cuda_inner_points);
	cudaFree(cudaRandXs);
	cudaFree(cudaRandYs);
	free(randXs);
	free(randYs);

	system("PAUSE");
	return 0;
}